-----------------------------------------------------------------------------------------------
-- Client Lua Script for Max_Loadouts
-- Copyright (c) NCsoft. All rights reserved
-----------------------------------------------------------------------------------------------
require "Window"
require "Apollo"
require "GameLib"
require "AbilityBook" 

-----------------------------------------------------------------------------------------------
-- Max_Loadouts Module Definition
-----------------------------------------------------------------------------------------------
local Max_Loadouts = {} 

-----------------------------------------------------------------------------------------------
-- Constants
-----------------------------------------------------------------------------------------------
local ML_settings = {}
local MlCount = {0, 0, 0, 0}
local MlVer = 1.14
local MlSpec = 1
local MlViewSpec = 1
local MlDelay = {0, 0, 0, 0}
local MlEdit = {0, 0, 0, 0}
local MLrmqueue = 0
local MlCurrent = 0
local nMlMax = 0
local nAbilityTime = 0
local bReset = false
local bNew = false
ML_settings[0] = {}
ML_settings[0][1] = {0, 1}
-----------------------------------------------------------------------------------------------
-- Initialization
-----------------------------------------------------------------------------------------------
function Max_Loadouts:new(o)
    o = o or {}
    setmetatable(o, self)
    self.__index = self 
	o.tItems = {} -- keep track of all the list items
	o.wndSelectedListItem = nil -- keep track of which list item is currently selected
    return o
end

function Max_Loadouts:Init()
    Apollo.RegisterAddon(self)
end
 
-----------------------------------------------------------------------------------------------
-- Max_Loadouts OnLoad
-----------------------------------------------------------------------------------------------
function Max_Loadouts:OnLoad()
	self.xmlDoc = XmlDoc.CreateFromFile("Max_Loadouts.xml")
	self.xmlDoc:RegisterCallback("OnDocLoaded", self)
end

-----------------------------------------------------------------------------------------------
-- Max_Loadouts OnDocLoaded
-----------------------------------------------------------------------------------------------
function Max_Loadouts:OnDocLoaded()
	if self.xmlDoc ~= nil and self.xmlDoc:IsLoaded() then
		self.wndbtn = Apollo.LoadForm(self.xmlDoc, "BarButton", nil, self)
	    self.wndMain = Apollo.LoadForm(self.xmlDoc, "Max_LoadoutsForm", nil, self)
		self.wndPrompt = Apollo.LoadForm(self.xmlDoc, "LoadoutNameForm", nil, self)
		self.wndRemove = Apollo.LoadForm(self.xmlDoc, "ConfirmRemove", nil, self)
		self.wndSave = Apollo.LoadForm(self.xmlDoc, "SetAbilities", nil, self)
		if self.wndMain == nil then
			Apollo.AddAddonErrorText(self, "Could not load the main window for some reason.")
			return Apollo.AddonLoadStatus.LoadingError
		end
		
		-- item list
		self.wndItemList = self.wndMain:FindChild("ItemList")
		self.wndbtn:Show(true)
	    self.wndMain:Show(false, true)
		
		-- Register handlers for events, slash commands and timer, etc.
		Apollo.RegisterSlashCommand("ml", "OnMax_LoadoutsToggle", self)
		Apollo.RegisterSlashCommand("ml1", "OnMax_LoadoutsMacro", self)
		Apollo.RegisterSlashCommand("ml2", "OnMax_LoadoutsMacro", self)
		Apollo.RegisterSlashCommand("ml3", "OnMax_LoadoutsMacro", self)
		Apollo.RegisterSlashCommand("ml4", "OnMax_LoadoutsMacro", self)
		Apollo.RegisterSlashCommand("mlreset", "OnReset", self)
		
		
		-- Do additional Addon initialization here
		Apollo.RegisterEventHandler("SendVarToML", "OnAbilitysUpdated", self)
		Apollo.RegisterEventHandler("SpecChanged", 	"OnSpecChange", self)
		
		-- register our Addon so others can wait for it if they want
		--g_AddonsLoaded["Max_Loadouts"] = true
		
		-- Set Spec Variables
		ML_settings[0][2] = MlVer
		MlSpec = AbilityBook.GetCurrentSpec()
		MlViewSpec = MlSpec
		self.wndMain:FindChild("ASProfile"):SetText("Action Set: "..MlSpec)
		self.wndPrompt:FindChild("KeepName"):SetData(22)
		if ML_settings[0][3] then
			self.wndMain:SetAnchorOffsets(unpack(ML_settings[0][3]))
		end
		if ML_settings[0][4] then
			self.wndbtn:SetAnchorOffsets(unpack(ML_settings[0][4]))
		end
		return Apollo.AddonLoadStatus.Loaded
	end
end

-----------------------------------------------------------------------------------------------
-- Max_Loadouts Functions
-----------------------------------------------------------------------------------------------
-- Define general functions here
-- on SlashCommand "/mlreset"
function Max_Loadouts:OnReset()
	ML_settings = nil
	ML_settings = {}
	MlCount = {0, 0, 0, 0}
	MlSpec = AbilityBook.GetCurrentSpec()
	MlViewSpec = MlSpec
	MlDelay = {0, 0, 0, 0}
	MlEdit = {0, 0, 0, 0}
	MLrmqueue = 0
	MlCurrent = 0
	nMlMax = 0
	ML_settings[0] = {}
	ML_settings[0][1] = {0, 1}
	ML_settings[0][2] = MlVer
	bReset = true
	RequestReloadUI()
end

-- on timer
function Max_Loadouts:OnTimer()
	if MlDelay[2] == 1 then -- If the Delay trigger is active
		if os.clock() > MlDelay[1] then -- and our 1 second has elapsed.
			AbilityBook.ClearCachedLASUpdates() 
			self:OnClose(MlDelay[3]) -- Run Onclose with the loadout stored from the delay
			MlDelay[2] = 0 -- Close the flag so the funtion doesn't keep looping.
			self.timer = nil
		end
	end
end

function Max_Loadouts:OnAbilitysUpdated()
	 nAbilityTime = os.clock()
	self.timer2 = ApolloTimer.Create(1, true, "OnTimer2", self)
end

function Max_Loadouts:OnTimer2()
	if os.clock() > nAbilityTime then -- and our 1 second has elapsed.
		self.timer2 = nil
		self:OnReceivedVar()
	end
end

function Max_Loadouts:OnSave(eLevel)
    if eLevel ~= GameLib.CodeEnumAddonSaveLevel.Character then
        return nil
    end
	if not bReset then
		ML_settings[0][3] = {self.wndMain:GetAnchorOffsets()}
		ML_settings[0][4] = {self.wndbtn:GetAnchorOffsets()}
	    return ML_settings
	else
		return {reset = true}
	end
end

function Max_Loadouts:OnRestore(eLevel, tData)
	if tData then
		if tData[0] then
			if tData[0][2] < 1.1 then
				-- Resets settings on older versions.
			else
				if not tData["reset"] then
					ML_settings = tData
					for i = 1, 4 do
						if ML_settings[i] then
							MlCount[i] = #ML_settings[i]
						end
					end
					if tData[0][1] then
						MlCurrent = tData[0][1][1]
						ML_settings[0][1][1] = MlCurrent
						ML_settings[0][1][2] = tData[0][1][2]
					else
						ML_settings[0][1] = {0, 1}
					end
				end
			end
		end
	end
end

function Max_Loadouts:OnReceivedVar()
	if bNew or MlEdit[1] == 1 then -- Only Caputure if the Max Loadout New button is clicked
		local loadout = nil
		MlSpec = AbilityBook.GetCurrentSpec() -- Make sure we are working with the same spec we are in
		if MlEdit[1] == 1 and MlEdit[3] == MlSpec then -- Checking if we are in edit mode and that we are in the appropriate spec
			loadout = MlEdit[2] -- set the number of the loadout we are editing
		else
			MlCount[MlSpec] = MlCount[MlSpec] + 1 -- Look into our table for the appropriate spec and get the next loadout number
			loadout = MlCount[MlSpec] -- set the number of the loadout we are creating
		end
		
		-- Make sure our Spec exists in the settings table, if not, create it.
		if not ML_settings[MlSpec] then
			ML_settings[MlSpec] = {}
		end
		
		-- Check if our loadoud exists in the settings table, probably not unless editing, so create it.
		if not ML_settings[MlSpec][loadout] then
			ML_settings[MlSpec][loadout] = {}
			ML_settings[MlSpec][loadout][1] = {}
			ML_settings[MlSpec][loadout][2] = {}
		end
					
		-- Check if you're a class with different stances
		if GameLib.GetClassInnateAbilitySpells().nSpellCount > 1 then
			ML_settings[MlSpec][loadout][6] = GameLib.GetCurrentClassInnateAbilityIndex() -- Set Stance to Loadout.
		end

		self.wndPrompt:Show(true) -- Show the Window to set Loadout Name.

		if MlEdit[1] == 1 then
			self.wndPrompt:FindChild("KeepName"):Show(true)
		else
			self.wndPrompt:FindChild("KeepName"):Show(false)
		end
		bNew = false
	end
end

function Max_Loadouts:OnSpecChange(SpecIndex, specError)
	MlSpec = SpecIndex
	self.wndMain:FindChild("ASProfile"):SetText("Action Set: "..SpecIndex)
	self:PopulateItemList(MlSpec)
	MlEdit[1] = 0 -- Make sure to turn edit mode off if the Spec changes.
end

-- Hook into the Abilities addon and capture Tier upgrades and put them in a table.
local MlAbilities = Apollo.GetAddon("Abilities")
local OrgOnSetActivateClick  = MlAbilities.OnSetActivateClick 
MlAbilities.OnSetActivateClick = function(self, wndHandler, wndControl) -- Hook and replicate function from the Carbine Abilities Addon 
	Event_FireGenericEvent("SendVarToML", "arSpellIds", "1")
	OrgOnSetActivateClick(self, wndHandler, wndControl)
end
-----------------------------------------------------------------------------------------------
-- Max_LoadoutsForm Functions
-----------------------------------------------------------------------------------------------
-- when the Cancel button is clicked
function Max_Loadouts:OnCancel()
	if self.wndMain:IsVisible()then
		self.wndMain:Show(false)
	end
	-- Make sure loadout editing is stopped if the window is closed.
	MlEdit[1] = 0
	bNew = false
end


function Max_Loadouts:OnLeftSet( wndHandler, wndControl, eMouseButton ) -- Back Arrow
	local maxSpec = AbilityBook.GetNumUnlockedSpecs() -- gets number of specs available to you.
	if MlViewSpec > 1 then -- If the spec (Action Set) you're viewing is higher than 1
		MlViewSpec = MlViewSpec - 1  --(e.g. Spec 3 moves to Spec 2)
	else
		MlViewSpec = maxSpec -- You're Looking at Action Set: 1, so the back arrow will take you to maxspecs
	end
	self:PopulateItemList(MlViewSpec) -- Rebuild ML Window for the new viewable Action Set
end

function Max_Loadouts:OnRightSet( wndHandler, wndControl, eMouseButton ) -- Forward Arrow
	local maxSpec = AbilityBook.GetNumUnlockedSpecs() -- gets number of specs available to you.
	if MlViewSpec == maxSpec then -- If you're at the maxspec moving forward should bring you to the first spec.
		MlViewSpec = 1
	else
		MlViewSpec = MlViewSpec + 1 --(e.g. Spec 2 moves to Spec 4)
	end
	self:PopulateItemList(MlViewSpec) -- Rebuild ML Window for the new viewable Action Set.
end


function Max_Loadouts:OnOpenAbilities( wndHandler, wndControl, eMouseButton )
	Event_FireGenericEvent("ToggleAbilitiesWindow")
	bNew = true
end

-----------------------------------------------------------------------------------------------
-- ItemList Functions
-----------------------------------------------------------------------------------------------
-- populate item list from the spec sent as the argument (spec)
function Max_Loadouts:PopulateItemList(spec)
	MlViewSpec = spec -- Make sure the Viewable Spec is the same sent in the argument.
	self.wndMain:FindChild("ASProfile"):SetText("Action Set: "..spec) -- Assign the Action Set number.
	self:DestroyItemList() -- make sure the item list is empty to start with.
	
	if MlCount[spec] ~= nil then -- Make sure the table of loadouts for the assigned spec exists.
		for i = 1,MlCount[spec] do -- Go through the table for the assigned spec and add each loadout.
			self:AddItem(i, spec)
		end
	end
	-- now all the item are added, call ArrangeChildrenVert to list out the list items vertically.
	self.wndItemList:ArrangeChildrenVert()
	MlEdit[1] = 0 -- Disable Editing when this function is called, to help prevent errors if a user clicks any arrows or trys anything funny when trying to edit a profile.
end

-- clear the item list
function Max_Loadouts:DestroyItemList()
	-- destroy all the wnd inside the list.
	for idx,wnd in ipairs(self.tItems) do
		wnd:Destroy()
	end

	-- clear the list item array.
	self.tItems = {}
	self.wndSelectedListItem = nil
end

-- add an item into the item list for the appropriate spec.
function Max_Loadouts:AddItem(loadout, spec)
	local wnd = Apollo.LoadForm(self.xmlDoc, "ListItem", self.wndItemList, self)
	
	-- keep track of the window item created.
	self.tItems[loadout] = wnd

	-- give it a piece of data to refer to.
	local wndItemText = wnd:FindChild("Text")
	if wndItemText then -- make sure the text wnd exist.
		if ML_settings[spec][loadout][4] ~= nil then -- If a loadout name exists.
			wndItemText:SetText(ML_settings[spec][loadout][4]) -- set the loadout name.
			if MlSpec == spec then
				if MlCurrent == loadout then
					wndItemText:SetTextColor(ApolloColor.new("White"))
				end
			end
		else -- Otherwise give it a generic name.
			wndItemText:SetText("Loadout: " .. loadout) 
		end
	end
	
	wnd:SetData(loadout) -- Add loadout number to wnd.
end

-- when a list item is selected.
function Max_Loadouts:OnListItemSelected(wndHandler, wndControl, mySpec, myLoadout)
    -- make sure the wndControl is valid.
	if wndHandler ~= wndControl then
        return
   	end
	
	if GameLib.GetPlayerUnit():IsInCombat() then
		Print("Cannot switch in combat")
		return
	else
		local loadout = 1
		self.wndSave:Show(true) -- Display the "Applying Settings" window.
		if mySpec > 0 then
			loadout = myLoadout
			MlSpec = mySpec
		else
			self.wndItemList:ArrangeChildrenVert()
			loadout = wndControl:GetData() -- Get the Loadout number associated with item that was selected.
			MlSpec = MlViewSpec -- Make sure the spec about to be processed matches the spec (Action Set) we are currently viewing. In case Event misfires
		end
		if MlSpec ~= AbilityBook.GetCurrentSpec() then -- No need to change if we're already in the right spec, prevents errors when messing around in the ability book
			AbilityBook.SetCurrentSpec(MlSpec) -- Set the spec of loadout we just selected.
		end

		
		if GameLib.GetClassInnateAbilitySpells().nSpellCount > 1 then -- If class has a stance, switch to the stance associated with the loadout
			GameLib.SetCurrentClassInnateAbilityIndex(ML_settings[MlSpec][loadout][6])
		end
	
		self.wndMain:Show(false) -- hide the window
 
		MlCurrent = loadout
		--ActionSetLib.RequestActionSetChanges(ML_settings[1][2][1]) -- submit ability index, this changes your action bar.
		-- Not sure this works, but it should be a safety net to make sure there are no pending changes to the LAS
		self.timer = ApolloTimer.Create(1, true, "OnTimer", self)
		-- Thanks to an internal cooldown on setting the specs, I have to go through a timer
		MlDelay[1] = os.clock() + 1 -- Give us 1 second
		MlDelay[2] = 1 -- This is a flag the timer knows to activate
		MlDelay[3] = loadout -- store the loadout name
		MlDelay[4] = MlSpec -- store the spec
	end
end

---------------------------------------------------------------------------------------------------
-- ListItem Functions
---------------------------------------------------------------------------------------------------
-- Delete a Loadout
function Max_Loadouts:OnRemoveListItem( wndHandler, wndControl, eMouseButton )
	MLrmqueue = wndControl:GetParent():GetData() -- Sets the remove queue number read from the item number associated with the delete button pressed
	self.wndRemove:Show(true)
end

-- ML Button pressed or SlashCommand "/ml"
function Max_Loadouts:OnMax_LoadoutsToggle()
	local ao1, ao2, ao3, ao4 = self.wndMain:GetAnchorOffsets()
	nMlMax = 0
	for i = 1, 4 do
		if ML_settings[i] then
			if nMlMax < #ML_settings[i] then
				nMlMax = #ML_settings[i]
			end
		end
	end
	
	ao4 = ao2 + 120 + (nMlMax * 35)
	
	self.wndMain:SetAnchorOffsets(ao1, ao2, ao3, ao4)
	self.wndMain:Show(true) -- show the window
	self:PopulateItemList(AbilityBook.GetCurrentSpec()) -- Make sure we build display the list of loadouts for our current spec
end

function Max_Loadouts:OnMax_LoadoutsMacro(msg, arg1)
	local loadout = tonumber(arg1)
	if msg == "ml1" then
		if ML_settings[1] and ML_settings[1][loadout] then
			self:OnListItemSelected(1, 1, 1, loadout)
		end
	elseif msg == "ml2" then
		if ML_settings[2] and ML_settings[2][loadout] then
			self:OnListItemSelected(1, 1, 2, loadout)
		end
	elseif msg == "ml3" then
		if ML_settings[3] and ML_settings[3][loadout] then
			self:OnListItemSelected(1, 1, 3, loadout)
		end
	elseif msg == "ml4" then
		if ML_settings[4] and ML_settings[4][loadout] then
			self:OnListItemSelected(1, 1, 4, loadout)
		end
	end
end
-- Applying Ability Tier Upgrades, called from the timer.
function Max_Loadouts:OnClose(loadout)
	-- Lets clear the ability points on everything we have loaded just to make sure we start fresh.
		local tAllAbilities = AbilityBook.GetAbilitiesList()
		for idx, tBaseAbility in pairs(tAllAbilities) do
			if tBaseAbility.nCurrentTier > 1 then
				AbilityBook.UpdateSpellTier(tBaseAbility.nId, 1)
			end
		end
		local currentAS = ActionSetLib.GetCurrentActionSet()	
		ML_settings[MlSpec][loadout][1][9] = currentAS[9] --Keep your gadget to prevent errors
	for i = 1, #ML_settings[MlSpec][loadout][2] do
		AbilityBook.UpdateSpellTier(ML_settings[MlSpec][loadout][2][i][1],ML_settings[MlSpec][loadout][2][i][2]) -- Apply the spell ID and Tier stored from the settings of the loadout.
	end
	
	ActionSetLib.RequestActionSetChanges(ML_settings[MlSpec][loadout][1])	
	ML_settings[0][1][1] = loadout
	ML_settings[0][1][2] = MlSpec
	self.wndSave:Show(false) -- Close the "Applying Settings" Window
end

-- Clicked the edit button
function Max_Loadouts:OnEdit( wndHandler, wndControl, eMouseButton )
	local loadout = wndControl:GetParent():GetData() -- get the ID of the loadout associated with the edit button
	if ML_settings[0][1][2] == MlViewSpec and MlCurrent == loadout then -- Make sure we are in the current loadout to enable editing, and we are in the right spec
		MlEdit[1] = 1 -- Activate Editing
		-- Make sure that we always edit ONLY this, just in case user changes Specs or Action Set Windows after editing.
		MlEdit[2] = loadout -- Store our loadout 
		MlEdit[3] = MlViewSpec -- Store our spec
		Event_FireGenericEvent("ToggleAbilitiesWindow")
	else
		Print("Max Loadout: Error - Your are not in the loadout you attempted to edit") -- As the error reads, they weren't in the loadout when they tried to edit it.
	end
end


 
function Max_Loadouts:OnCreateMacro( wndHandler, wndControl, eMouseButton )
	local loadout = wndControl:GetParent():GetData()
	local strName = "ML Macro Spec:"..MlViewSpec..", Loadout:"..loadout
	local strSprite = "IconSprites:Icon_Modifier_endurance_max_001"
	local strCmds = "/ml"..MlViewSpec.." "..loadout
	local id = MacrosLib.CreateMacro()
	MacrosLib.SetMacroData( id, true, strName, strSprite, strCmds )
	Event_FireGenericEvent("InterfaceMenu_ToggleMacro")
end

---------------------------------------------------------------------------------------------------
-- LoadoutNameForm Functions
---------------------------------------------------------------------------------------------------
-- Name the loadout box
function Max_Loadouts:OnLoadoutName( wndHandler, wndControl )
	self.wndPrompt:Show(false) -- hide the Text Box
	ldoutname = self.wndPrompt:FindChild("LoadoutEditBox"):GetText() -- Grab what was typed
	if MlEdit[1] == 1 then -- was this an edit?
		MlCnt = MlEdit[2] -- set the loadout we stored in edit
		if wndControl:GetData() ~= 22 then
			ML_settings[MlSpec][MlCnt][4] = ldoutname -- Set the name
		end
		MlEdit[1] = 0 -- edit mode off
	else -- this is a new entry
		MlCnt = MlCount[MlSpec] -- The current tally of loadouts for this spec
		ML_settings[MlSpec][MlCnt][4] = ldoutname -- Set the name
		MlCurrent = MlCnt
	end
	self.wndMain:Show(false) -- hide the window
	self.wndPrompt:FindChild("LoadoutEditBox"):SetText("") -- Clear the text in the form for next use.
	local tAllAbilities = AbilityBook.GetAbilitiesList()
	local currentAS = ActionSetLib.GetCurrentActionSet()
	for i=1, 12 do
		for idx, tBaseAbility in pairs(tAllAbilities) do
			if tBaseAbility.nId == currentAS[i] then
				ML_settings[MlSpec][MlCnt][1][i] = currentAS[i]
				if i < 9 then
					ML_settings[MlSpec][MlCnt][2][i] = {}
					ML_settings[MlSpec][MlCnt][2][i][1] = currentAS[i]
					ML_settings[MlSpec][MlCnt][2][i][2] = tBaseAbility.nCurrentTier
				end
				break
			else
				ML_settings[MlSpec][MlCnt][1][i] = 0
			end
		end
	end
end

---------------------------------------------------------------------------------------------------
-- ConfirmRemove Functions
---------------------------------------------------------------------------------------------------
function Max_Loadouts:OnRemoveLoadout( wndHandler, wndControl )
	self.wndRemove:Show(false) -- hide the Remove Box
	local loadout = MLrmqueue -- Sets the loudout based on what's in the remove queue
	local ldoutCnt = #ML_settings[MlViewSpec] -- Sets the number of loadouts currently in our settings associated with the spec
	if loadout >= ldoutCnt then -- Is it the last loadout on the list?
		ML_settings[MlViewSpec][loadout] = nil -- Simply remove the last loadout
	else
		local ldoutTot = ldoutCnt - loadout -- Find out the difference between them. If there are 5 loadouts, and loadout 2 is selected to be deleted, 5-2=ldoutTot(3)
		for i = 1, ldoutTot do -- Now we need to have the loadouts that were below the deleted item to shift up (ldoutTot=3 from the example above)
			local ldoutKil = i + loadout -- Get the loadouts after the one selected for deletion. (If ldoutTot=3, and 2 was selected, we would get here, 3, 4 and then 5)
			ML_settings[MlViewSpec][ldoutKil - 1] = ML_settings[MlViewSpec][ldoutKil] -- bumps each of them up a number (with example above, 2=3, 3=4, 4=5)
		end
		ML_settings[MlViewSpec][ldoutCnt] = nil -- And then kill the final (5 in the example above.)
	end
	MlCount[MlViewSpec] = MlCount[MlViewSpec] - 1 -- Make sure our counter knows there is one less
	self:PopulateItemList(MlViewSpec) -- Refresh the window with the new list of loadouts
end

function Max_Loadouts:OnRemoveLoadoutCancel( wndHandler, wndControl, eMouseButton )
	self.wndRemove:Show(false) -- hide the Remove Box
	MLrmqueue = 0
end

-----------------------------------------------------------------------------------------------
-- Max_Loadouts Instance
-----------------------------------------------------------------------------------------------
local Max_LoadoutsInst = Max_Loadouts:new()
Max_LoadoutsInst:Init()
